
package com.ibm.oti.vm;

/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 1998, 2010  All Rights Reserved
 */

import java.util.*;
import java.io.FilePermission;

/**
 * BootstrapClassLoaders load classes out of the file system,
 * from the directories and jars on the system class path.
 * The system class path can be set using the command line
 * option "-scp <paths>", and can be read from the system
 * property "sun.boot.class.path".
 *
 * @author		OTI
 * @version		initial
 */

public final class BootstrapClassLoader extends AbstractClassLoader{

	static {
		com.ibm.oti.vm.VM.initializeVM();
		Class cl = Integer.class;
	}
	private static BootstrapClassLoader singleton;
	private Hashtable packages = new Hashtable();
	ThreadLocal definingPackage = new ThreadLocal();

/**
 * Prevents this class from being instantiated.
 * BootstrapClassLoaders may only be created by the VM.
 *
 * @author		OTI
 * @version		initial
 */
private BootstrapClassLoader() {
	int count = VM.getClassPathCount();
	types = new int[count];
	cache = new Object[count];
	parsedPath = new String[count];
	VM.initializeClassLoader(this, true);
}

/**
 * Invoked by the Virtual Machine when resolving class references.
 * Equivalent to loadClass(className, false);
 *
 * @return 		java.lang.Class
 *					the Class object.
 * @param 		className String
 *					the name of the class to search for.
 *
 * @exception	ClassNotFoundException
 *					If the class could not be found.
 */
public Class loadClass(String className) throws ClassNotFoundException {
	Class loadedClass;
	loadedClass = com.ibm.oti.vm.VM.findClassOrNull(className, this);

    if (loadedClass != null) {
        String packageName = getPackageName(loadedClass);
        if (packageName != null && super.getPackage(packageName) == null) {
            int index = VM.getCPIndexImpl(loadedClass);
                addPackage(packageName, index);
        }
    }
	return loadedClass;
}

private void addPackage(String packageName, int index) {
    synchronized(packages) {
        if (!packages.containsKey(packageName)) {
            packages.put(packageName, new Integer(index));
        }
    }
}

public static ClassLoader singleton() {
	if (singleton == null)
	 	singleton = new BootstrapClassLoader();
	 else
	 	// K0084 = can only instantiate one BootstrapClassLoader
	 	throw new SecurityException(com.ibm.oti.util.Msg.getString("K0084"));
 	return singleton;
}

protected Package getPackage(String name) {
	Package result;
	Integer index = (Integer)packages.get(name);
	if (index == null) {
		return super.getPackage(name);
	}
	String inPackage = (String)definingPackage.get();
	if (name.equals(inPackage)) return null;
	if (inPackage != null) throw new InternalError();
	definingPackage.set(name);
	try {
		result = definePackage(name, index.intValue());
		packages.remove(name);
	} finally {
		definingPackage.set(null);
	}
	return result;
}

protected Package[] getPackages() {
	// Capture packages loaded while calling definePackage().
	// Also ensure all Hashtable elements are defined since the
	// Hashtable is being modified by getPackage() while its
	// being walked
	while (packages.size() > 0) {
		Hashtable packagesClone = (Hashtable)packages.clone();
		Enumeration keys = packagesClone.keys();
		while (keys.hasMoreElements()) {
			getPackage((String)keys.nextElement());
		}
	}
	return super.getPackages();
}

/**
 * Invoked by the Virtual Machine when a JVMTI agent wishes to add to the bootstrap class path.
 * The VM has already verified that the jar/zip exists and is superficially valid.
 *
 * @param 		className jarPath
 *					the name of the jar/zip to add.
 */
private void appendToClassPathForInstrumentation(String jarPath) throws Throwable {
	synchronized(cacheLock) {
		// Note: No java-level synchronization is required to call the native.
		int newCount = addJar(com.ibm.oti.util.Util.getBytes(jarPath));

		int[] newTypes = new int[newCount];
		System.arraycopy(types, 0, newTypes, 0, newCount - 1);
		types = newTypes;

		Object[] newCache = new Object[newCount];
		System.arraycopy(cache, 0, newCache, 0, newCount - 1);
		cache = newCache;

		String[] newParsedPath = new String[newCount];
		System.arraycopy(parsedPath, 0, newParsedPath, 0, newCount - 1);
		parsedPath = newParsedPath;

		if (permissions != null) {
			FilePermission[] newPermissions = new FilePermission[newCount];
			System.arraycopy(permissions, 0, newPermissions, 0, newCount -1);
			permissions = newPermissions;
		}
	}
}

/**
 * Add a JAR/ZIP to the boot class path, creating a new class path entry for it.
 *
 * @return 		int
 *					the new size othe class path entry array
 * @param 		jarPath byte[]
 *					the name of the jar/zip to add, in OS filename encoding
 */
private native int addJar(byte jarPath[]);

}
